import 'dart:async';

import 'package:flutter/foundation.dart';
import 'package:osam/domain/event/event.dart';
import 'package:osam/domain/middleware/middleware.dart';
import 'package:osam/domain/state/base_state.dart';
import 'package:osam/persist/persist_repository.dart';

abstract class Store<ST extends BaseState<ST>> implements Persist {
  factory Store(ST state, {List<Middleware<ST>> middleWares = const [], bool logging = false}) =>
      _StoreImpl(appState: state, middleWares: middleWares, logging: logging);

  ST get state;

  bool get isPersistEnabled;

  Stream<Event<ST>> get eventStream;

  void updateMiddleware(Middleware<ST> middleware);

  void updateMiddlewares(List<Middleware<ST>> middlewares);

  void dispatchEvent({@required Event<ST> event});
}

class _StoreImpl<ST extends BaseState<ST>> implements Store<ST> {
  ST appState;
  final List<Middleware<ST>> middleWares;
  final _denormalizedConditions = <Condition<ST>>[];
  final bool logging;

  // ignore: close_sinks
  StreamController<Event<ST>> _dispatcher;

  _StoreImpl({this.appState, this.middleWares, this.logging}) {
    _initStore();
  }

  @override
  ST get state => appState;

  @override
  bool get isPersistEnabled => PersistRepository().isEnabled;

  @override
  Stream<Event<ST>> get eventStream => _dispatcher.stream;

  @override
  Future<void> initPersist(String appName) async => await PersistRepository().init(appName);

  @override
  void dispatchEvent({@required Event<ST> event}) => _dispatcher.sink.add(event);

  @override
  void storeState() => PersistRepository().storeState(appState);

  @override
  void restoreState() => this.appState = PersistRepository().restoreState() ?? this.appState;

  @override
  void deleteState() => PersistRepository().deleteState();

  void _denormalizeConditions() {
    _denormalizedConditions.clear();
    _denormalizedConditions
        .addAll(middleWares.expand((middleWare) => middleWare.conditions).toList());
  }

  void _setupStoreToMiddlewares(List<Middleware<ST>> middlewares) =>
      middlewares.forEach((middleware) => middleware.store = this);

  @override
  void updateMiddleware(Middleware<ST> middleware) {
    this.middleWares[this.middleWares.indexOf(middleware)] = middleware;
    _setupStoreToMiddlewares([middleware]);
    _denormalizeConditions();
  }

  @override
  void updateMiddlewares(List<Middleware<ST>> middlewares) {
    middlewares.forEach(
        (middleware) => this.middleWares[this.middleWares.indexOf(middleware)] = middleware);
    _setupStoreToMiddlewares(middlewares);
    _denormalizeConditions();
  }

  void _initStore() {
    _dispatcher = StreamController<Event<ST>>.broadcast();
    _setupStoreToMiddlewares(this.middleWares);
    _denormalizeConditions();
    _dispatcher.stream.listen((event) {
      for (Condition<ST> condition in _denormalizedConditions) {
        final nextEvent = condition(event);
        if (!nextEvent) return;
      }
      if (logging)
        debugPrint(
            'Event in stores event stream is : ${'runtimeType: ' + event.runtimeType.toString()}');
      if (event.reducer != null) {
        try {
          event(appState);
        } catch (e) {
          print('store error while reducer calling: $e');
        }
      }
    });
  }
}
