import 'package:flutter/material.dart';
import 'package:osam/domain/state/base_state.dart';
import 'package:osam/domain/store/store.dart';
import 'package:osam/presentation/presenter.dart';
import 'package:provider/provider.dart';

class StoreProvider<S extends Store<BaseState>> extends StatelessWidget {
  final S store;
  final Widget child;
  final lifecycleProviderKey = GlobalKey();
  final storeProviderKey = GlobalKey();

  StoreProvider({Key key, @required this.store, @required this.child}) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return _InternalStoreProvider(
        key: storeProviderKey,
        store: store,
        child: store.isPersistEnabled
            ? _LifecycleWrapper<S>(child: child, key: lifecycleProviderKey)
            : child);
  }
}

class _InternalStoreProvider extends InheritedWidget {
  const _InternalStoreProvider({
    @required Key key,
    @required this.child,
    @required this.store,
  }) : super(key: key, child: child);

  final Store store;
  final Widget child;

  static _InternalStoreProvider of(BuildContext context) =>
      context.dependOnInheritedWidgetOfExactType(aspect: _InternalStoreProvider);

  @override
  bool updateShouldNotify(_InternalStoreProvider oldWidget) => this.store != oldWidget.store;
}

class _LifecycleWrapper<S extends Store<BaseState>> extends StatefulWidget {
  final Widget child;

  const _LifecycleWrapper({Key key, this.child}) : super(key: key);

  @override
  _LifecycleWrapperState createState() => _LifecycleWrapperState<S>();
}

class _LifecycleWrapperState<S extends Store<BaseState>> extends State<_LifecycleWrapper>
    with WidgetsBindingObserver {
  @override
  void initState() {
    WidgetsBinding.instance.addObserver(this);
    super.initState();
  }

  @override
  void dispose() {
    WidgetsBinding.instance.removeObserver(this);
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return widget.child;
  }

  @override
  void didChangeAppLifecycleState(AppLifecycleState state) {
    if (state == AppLifecycleState.paused ||
        state == AppLifecycleState.inactive ||
        state == AppLifecycleState.detached) _InternalStoreProvider.of(context).store.storeState();
  }
}

class PresenterProvider<P extends Presenter> extends StatelessWidget {
  final P presenter;
  final Widget child;

  const PresenterProvider({Key key, @required this.presenter, @required this.child})
      : super(key: key);

  @override
  Widget build(BuildContext context) {
    final store = _InternalStoreProvider.of(context).store;
    assert(store != null);
    if (store == null)
      debugPrint('insure you provided store by StoreProvider in the start of your widget tree');
    return Provider(
      key: ValueKey('PresenterProvider: %${key.toString()}'),
      create: (ctx) => presenter
        // ignore: invalid_use_of_protected_member
        ..store = store,
      child: child,
      dispose: (ctx, presenter) => presenter.dispose(),
    );
  }

  static P of<P extends Presenter>(BuildContext context) => Provider.of<P>(context, listen: false);
}
